// UpgradeBuilder.cpp : Build empeg/RioCar upgrade file
//

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>
#ifdef WIN32
#include <io.h>
#endif
#include "crc.h"
#include "defines.h"

#define BUFSIZE		4096

#ifdef WIN32
#define STAT _stat
#define OFF_T _off_t
#else
#define _S_IFREG S_IFREG
#define OFF_T __off_t
#define STAT stat
#endif

typedef struct chunktag {
	unsigned long	typ;
	unsigned long	length;
	unsigned long	addr;	// only valid for FLASH* chunks
	char*			fname;
} chunk;

typedef struct configtag {
	// the control input file
	char*	ctlname;
	// the temporary output file
	char*	tmpname;
	// final output file
	char*	outname;
	// show commandline help?
	bool	show_help;
	// info for upgrade file
	unsigned long	upgraderversion;
	long			upgrade_size;
	char*			info;
	char*			what;
	char*			release;
	char*			version;
	char*			hwrev;
	chunk**			chnk;
	unsigned char	nchunks;
	unsigned char	maxchunks;
	bool			upgver_read;
	bool			info_read;
	bool			what_read;
	bool			rel_read;
	bool			ver_read;
	bool			doit;
} config;

// some help definitions/macros
//
// print a chunk
#define PRINTCHUNK(x)	//
/* #define PRINTCHUNK(x)  {	fprintf(stdout,"Found chunk type: 0x%2.2X\n",x->typ); fprintf(stdout,"Filename: %s\n",x->fname); fprintf(stdout,"address: 0x%8.8X\n",x->addr); };			*/
#define ALLOCATECHUNK(x) { x=(chunk*)malloc(sizeof(chunk)); if (!x) { fprintf(stderr,"ERROR: out of memory allocating chunk storage\n"); return true; }; tchnk->fname=(char*)NULL; tchnk->addr=0; tchnk->length=0; };
#define PRINTERRCODE()	fprintf(stderr,"Errorcode: %s\n",errstr(errno));
// end of help definitions/macros

// initialize cfg with default values
bool init_cfg(config* cfg);

// parse commandline args
bool parse_args(const int argc, char* argv[], config* cfg);

// show commandline help
void show_help(const char* exename);

// read the config file
bool read_cfg(config *cfg);

// check the config for consistency
bool check_cfg(config *cfg);

// write header and chunks
bool write_file(const config *cfg);

// calculate and append CRC32
bool do_crc32(const config *cfg);

// append file source to file target, both files should already be open
bool appendfile(FILE* target, char* sourcename);

// return clear text string for error errnumber
const char* errstr(const int errnumber);

// write an unsigned long in the right order
bool writeUL(FILE* f, const unsigned long ulong);

int main(int argc, char* argv[])
{
	config	cfg;

	if (init_cfg(&cfg)) return UB_IC_ERR;

	if (parse_args(argc, argv, &cfg)) return UB_ARG_ERR;

	if (cfg.show_help) {
		show_help(argv[0]);
		return UB_HELP_SHOWN;
	};

	if (read_cfg(&cfg)) return UB_CFG_ERR;

	if (check_cfg(&cfg)) return UB_CHK_ERR;

	// are we asked to actually write files?
	// if not, everything is done
	if (!cfg.doit) return UB_NO_ERR;

	if (write_file(&cfg)) return UB_TWF_ERR;

	if (do_crc32(&cfg)) return UB_TWC_ERR;

	return UB_NO_ERR;
};

//******************************
//** function implementations **
//******************************

bool init_cfg(config* cfg)
{
	cfg->ctlname="upgrade.ctl";
	cfg->tmpname="upg_tempfile";
	cfg->outname="output.upgrade";
	cfg->show_help=false;
	cfg->upgraderversion=UPGRADERVERSION;
	cfg->info="Build by UpgradeBuilder (c)2001 Sven Mueller <sven@incase.de>\n";
	cfg->what="an upgrade file\n";
	cfg->release="\n";
	cfg->version="\n";
	cfg->hwrev="\n";
	cfg->chnk=(chunk**)malloc(128*sizeof(chunk*));
	cfg->nchunks=0;
	if (cfg->chnk) {
		cfg->maxchunks=128;
	} else {
		fprintf(stderr,"ERROR: Out of memory while initializing config structure\n");
		return true;
	};
	cfg->upgver_read=false;
	cfg->info_read=false;
	cfg->what_read=false;
	cfg->rel_read=false;
	cfg->ver_read=false;
	cfg->doit=true;
	return false;
};

bool parse_args(const int argc, char* argv[], config* cfg) {
	int curarg=1;

	while ((curarg < argc) && (!cfg->show_help)) {
		if (argv[curarg][0]=='-') {
			if (strlen(argv[curarg])>2) return true;
			switch(argv[curarg][1]) {
				case 'h':
				case 'H':	cfg->show_help=true;
							break;
				case 'n':
				case 'N':	cfg->doit=false;
							break;
				case 'c':
				case 'C':	if (++curarg > argc) {
								fprintf(stderr,"ERROR: Argument missing for -c option\n");
								cfg->show_help=true;
								return true;
							};
							if (argv[curarg][0]=='-') {
								fprintf(stderr,"ERROR: Argument missing for -c option\n");
								curarg--;
								cfg->show_help=true;
								return true;
							};
							cfg->ctlname=strdup(argv[curarg]);
							break;
				case 'o':
				case 'O':	if (++curarg < argc) {
								if (argv[curarg][0]=='-') {
									fprintf(stderr,"ERROR: Argument missing for -o option\n");
									curarg--;
									cfg->show_help=true;
									return true;
								};
								strcpy(cfg->outname,argv[curarg]);
							} else {
								fprintf(stderr,"ERROR: Argument missing for -o option\n");
								cfg->show_help=true;
								return true;
							};
							break;
				default:	fprintf(stderr,"ERROR: Unknown command line option %s\n",argv[curarg]);
							cfg->show_help=true;
							return true;
			};
		} else {
			fprintf(stderr,"ERROR: unexpected non-option parameter <%s>\n",argv[curarg]);
			cfg->show_help=true;
			return true;
		};
		curarg++;
	};
	return false;
};

void show_help(const char* exename) {
	fprintf(stdout,"Usage:\n");
	fprintf(stdout,"%s [-h] [-n] [-o <outputfile>] [-c <controlfile>]\n",exename);
	fprintf(stdout,"    -h              prints this help\n");
	fprintf(stdout,"    -n              don't write any files, just check config\n");
	fprintf(stdout,"    -o <outputfile> sets the name of the output (upgrade) file\n");
	fprintf(stdout,"                    defaults to \"output.upgrade\"\n");
	fprintf(stdout,"    -c <configfile> sets the name of the input (config) file\n");
	fprintf(stdout,"                    defaults to \"upgrade.ctl\"\n");
	fprintf(stdout,"The upgrade file is created from various file, with details being\n");
	fprintf(stdout,"read from the config file\n");
	return;
};

const char* errstr(const int errnumber){
	static char	buf[256];

	switch (errnumber) {
		case	ENOENT:		return("No such file or directory"); 
							break;
		case	E2BIG:		return("Argument list too long");
							break;
		case	ENOEXEC:	return("Exec format error");
							break;
		case	EBADF:		return("Bad file number");
							break;
		case	ECHILD:		return("No spawned processes");
							break;
		case	EAGAIN:		return("No more processes or not enough memory or maximum nesting level reached");
							break;
		case	ENOMEM:		return("Not enough memory");
							break;
		case	EACCES:		return("Permission denied");  
							break;
		case	EEXIST:		return("File exists");
							break;
		case	EXDEV:		return("Cross-device link");
							break;
		case	EINVAL:		return("Invalid argument");
							break;
		case	EMFILE:		return("Too many open files");
							break;
		case	ENOSPC:		return("No space left on device");
							break;
		case	EDOM:		return("Math argument");
							break;
		case	ERANGE:		return("Result too large");
							break;
		case	EDEADLOCK:	return("Resource deadlock would occur");
							break;
		default:			sprintf(buf,"unknown error code: %d",errnumber);
							return(buf);
	}
};

bool read_cfg(config *cfg) {
	
	char*	inbuf=(char*)NULL;
	char	command[256];
	char	option[256];
	char*	eqpos;
	chunk*	tchnk;
	int		tmp_errno;
	FILE*	ctlfile;
//	char*	temp;

	ctlfile=fopen(cfg->ctlname,"rt");
	if (!ctlfile) {
		fprintf(stderr,"ERROR: Could not open controlfile <%s>\n",cfg->ctlname);
		PRINTERRCODE();
	};

	// read the whole file
	while (!feof(ctlfile)) {
		// clean up buffers
		memchr(command,0,256);
		memchr(option,0,256);
		if (inbuf) free(inbuf);
		inbuf=(char*)malloc(256);
		inbuf[255]=0;
		if (!inbuf) {
			fprintf(stderr,"ERROR: Out of memory while reading control file\n");
			return true;
		};
		// read a line
		if (feof(ctlfile)) fprintf(stderr,"End of file\n");
		if (!fgets(inbuf,255,ctlfile)) {
			tmp_errno=errno;
			if (feof(ctlfile)) {
				// read failed because we reached EOF, this isn't really an error
				return false;
			};
			fprintf(stderr,"ERROR: Could not read from control file\n");
			fprintf(stderr,"Error code: %s\n",errstr(tmp_errno));
			return true;
		};
		if (strlen(inbuf)>254) {
			fprintf(stderr,"ERROR: Control line too long: %s\n",inbuf);
			return true;
		};
		while ((inbuf[strlen(inbuf)-1]==0x0A) || (inbuf[strlen(inbuf)-1]==0x0D))
			inbuf[strlen(inbuf)-1]=0;
		fprintf(stdout,"line retrieved: %s\n",inbuf);
		// parse the line (split it into its parts <command>=<option>)
		eqpos=strchr(inbuf,'=');
		if (!eqpos) {
			if ((strncmp(inbuf,"pumphd",6)==0) && (strlen(inbuf)==7)){
				ALLOCATECHUNK(tchnk);
				tchnk->typ=0x20+(0x10*(inbuf[6]-'a'));
				if ((tchnk->typ<0x20) || (tchnk->typ>0x48) || ((tchnk->typ)&0xf >8)) {
					fprintf(stderr,"ERROR: unknown partition %#x (%c) for command: %s\n",
						tchnk->typ,
						((tchnk->typ & 0x70)>>4)-2+'a',
						inbuf);
					return true;
				};
				cfg->chnk[cfg->nchunks++]=tchnk;
				PRINTCHUNK(tchnk);
			} else {
				fprintf(stderr,"ERROR: Format error while trying to parse line:\n%s\n",inbuf);
				return true;
			};
		} else {
			strncpy(command,inbuf,eqpos-inbuf);
			command[eqpos-inbuf]=0;
			eqpos++;
			strcpy(option,eqpos);
//			fprintf(stdout,"Split it into command >%s< option >%s<\n",command,option);
			if (strcmp(command,"upgraderversion")==0) {
				if (sscanf(option,"%d",&(cfg->upgraderversion))!=1) {
					fprintf(stderr,"ERROR: Could not parse option for upgraderversion command: %s\n",option);
					return true;
				};
			} else if (strcmp(command,"infofile")==0) {
				ALLOCATECHUNK(tchnk);
				tchnk->fname=strdup(option);
				tchnk->typ=CHUNK_INFO;
				tchnk->length=0; // this gets filled in b< check_cfg
				cfg->chnk[cfg->nchunks++]=tchnk;
				PRINTCHUNK(tchnk);
			} else if (strcmp(command,"whatfile")==0) {
				ALLOCATECHUNK(tchnk);
				tchnk->fname=strdup(option);
				tchnk->typ=CHUNK_WHAT;
				tchnk->length=0; // this gets filled in b< check_cfg
				cfg->chnk[cfg->nchunks++]=tchnk;
				PRINTCHUNK(tchnk);
			} else if (strcmp(command,"releasefile")==0) {
				ALLOCATECHUNK(tchnk);
				tchnk->fname=strdup(option);
				tchnk->typ=CHUNK_RELEASE;
				tchnk->length=0; // this gets filled in b< check_cfg
				cfg->chnk[cfg->nchunks++]=tchnk;
				PRINTCHUNK(tchnk);
			} else if (strcmp(command,"versionfile")==0) {
				ALLOCATECHUNK(tchnk);
				tchnk->fname=strdup(option);
				tchnk->typ=CHUNK_VERSION;
				tchnk->length=0; // this gets filled in b< check_cfg
				cfg->chnk[cfg->nchunks++]=tchnk;
				PRINTCHUNK(tchnk);
			} else if (strcmp(command,"hwrevfile")==0) {
				ALLOCATECHUNK(tchnk);
				tchnk->fname=strdup(option);
				tchnk->typ=CHUNK_HWREV;
				tchnk->length=0; // this gets filled in b< check_cfg
				cfg->chnk[cfg->nchunks++]=tchnk;
				PRINTCHUNK(tchnk);
			} else if (strncmp(command,"flash",5)==0) {
				ALLOCATECHUNK(tchnk);
				if (strcmp(&(command[5]),"loader")==0) {
					tchnk->typ=CHUNK_FLASHLOADER;
				} else if (strcmp(&(command[5]),"kernel")==0) {
					tchnk->typ=CHUNK_FLASHKERNEL;
				} else if (strcmp(&(command[5]),"ramdisk")==0) {
					tchnk->typ=CHUNK_FLASHRAMDISK;
				} else if (strcmp(&(command[5]),"random")==0) {
					tchnk->typ=CHUNK_FLASHRANDOM;
				} else {
					fprintf(stderr,"ERROR: unknown flash command: %s\n",command);
					return true;
				};
				memchr(inbuf,0,255);
				if (sscanf(option,"0x%x,%s",&(tchnk->addr),inbuf)!=2) {
					fprintf(stderr,"ERROR: Could not parse flashcommands optionstring: >%s<\n",option);
					return true;
				};
				tchnk->fname=(char*)malloc(strlen(inbuf)+1);
				if (!tchnk->fname) {
					fprintf(stderr,"ERROR: out of memory allocating memory for chunk filename\n");
					return true;
				};
				memcpy(tchnk->fname,inbuf,strlen(inbuf)+1);
				tchnk->fname[strlen(inbuf)]=0; // make real sure the string is terminated like it should be
				cfg->chnk[cfg->nchunks++]=tchnk;
				PRINTCHUNK(tchnk);
			} else if (strncmp(command,"pumphd",6)==0) {
				ALLOCATECHUNK(tchnk);
				tchnk->fname=strdup(option);
				tchnk->addr=0;
				tchnk->length=0;
				tchnk->typ=0x20+0x10*(command[6]-'a');
				if (strlen(command)==7) {
					// pump disk (initialize partition table)
					free(tchnk->fname);
					tchnk->fname=(char*)NULL;
				} else {
					tchnk->typ+=command[7]-'0';
				};
				if ((tchnk->typ < 0x20) || (tchnk->typ > 0x48) || (((tchnk->typ) & 0xf) > 8)) {
					fprintf(stderr,"ERROR: unknown partition (chunk type would be ");
					fprintf(stderr,"%#2.2x) for command: %s\n", tchnk->typ, command);
					return true;
				};
				cfg->chnk[cfg->nchunks++]=tchnk;
				PRINTCHUNK(tchnk);
			} else {
				fprintf(stderr,"ERROR: Unknown config file command:\n  %s\n",command);
				return true;
			};
		};
		if (cfg->nchunks >= cfg->maxchunks) {
			fprintf(stderr,"ERROR: Too many chunks in upgrade file: %d\n",cfg->nchunks);
			return true;
		};
	}; // 	while (!feof(cfg->ctlfile) && !stopnow)
	return false;
};

char* getfreename(const char* basname){
	char*	buf;
	char*	rbuf;
	struct STAT st;
	size_t	curlen;
	bool	exists;
	static char	*posext="0123456789abcdefghijklmnopqrstuvwxyz";
	int		numexts;
	int		curext=0;

	buf=(char*)malloc(256);
	if (!buf) {
		fprintf(stderr,"ERROR: out of memory trying to find a new extenson\n");
	};
	strncpy(buf,basname,255);
	buf[255]=0;
	curlen=strlen(buf);
	buf[curlen]=0;
	buf[curlen+1]=0;
	numexts=strlen(posext);
	if (STAT(buf,&st)!=0) {
		exists=(errno=ENOENT)?false:true;
	};

	
	// if file "basname" does not exist: return it as a free name
	if (!exists) return buf; 
	
	// first check wether a single character extension is enough
	while (exists && (curext<numexts)) {
		buf[curlen]=posext[curext];
		if (STAT(buf,&st)!=0) {
			exists=(errno=ENOENT)?false:true;
		};
		if (!exists) return buf;
		curext++;
	};

	curext=0;
	while (exists && (curext<numexts)) {
		buf[curlen]=posext[curext];
		rbuf=getfreename(buf);
		if (rbuf) {
			free(buf);
			return rbuf;
		};
		curext++;
	};
	free(buf);
	return((char*)NULL);
};

bool check_cfg(config *cfg) {
	char* tmp_n;
	struct STAT st;
	unsigned char	cur_idx1,cur_idx2;
	chunk*	tmp_ch;
	FILE*	tfile;
	unsigned long ctyp;
	chunk*	cchnk;

	cfg->upgrade_size=0;	// start at zero bytes upgrade size.

	tmp_n=getfreename(cfg->tmpname);
	if (!tmp_n) {
		fprintf(stderr,"ERROR: No free filename for temporary file found\n");
		return true;
	};
	cfg->tmpname=tmp_n;
	// final output file
	tmp_n=getfreename(cfg->outname);
	if (!tmp_n) {
		fprintf(stderr,"ERROR: No free filename for output file found\n");
		return true;
	};
	cfg->outname=tmp_n;

	// info for upgrade file
	if (cfg->upgraderversion != 0x02) {
		fprintf(stderr,"ERROR: unknown config file version 0x%X\n",cfg->upgraderversion);
		return true;
	};
	fprintf(stdout,"UpgraderVersion: 0x%2.2x\n",cfg->upgraderversion);

	cur_idx1=0;
	while (cur_idx1 < cfg->nchunks) {
		if ((cfg->chnk[cur_idx1]->typ == CHUNK_INFO)
		||  (cfg->chnk[cur_idx1]->typ == CHUNK_WHAT)
		||	(cfg->chnk[cur_idx1]->typ == CHUNK_RELEASE)
		||	(cfg->chnk[cur_idx1]->typ == CHUNK_VERSION)
		|| (cfg->chnk[cur_idx1]->typ == CHUNK_HWREV)) {
			tmp_ch=cfg->chnk[cur_idx1];
			if (cur_idx1 == (cfg->nchunks-1)) {
				(cfg->nchunks)--;
			} else {
				for (cur_idx2=cur_idx1; (cur_idx2 < (cfg->nchunks-1)); cur_idx2++)
					cfg->chnk[cur_idx2]= cfg->chnk[cur_idx2+1];
				(cfg->nchunks)--;
			};
			switch (tmp_ch->typ) {
				case CHUNK_INFO:
					cfg->info_read=true;
					if (STAT(tmp_ch->fname,&st)) {
						fprintf(stderr,"ERROR: Could not stat file %s\n",tmp_ch->fname);
						return true;
					};
					if (st.st_size==0)
						fprintf(stderr,"Do you really intend not to set the >info< tag?\n");
					cfg->info = (char*)(malloc(st.st_size+1));
					fprintf(stdout,"Info file: %s\t size: %d\n",tmp_ch->fname,st.st_size);
					memchr(cfg->info,0,st.st_size+1);
					if (!(cfg->info)) {
						fprintf(stderr,"ERROR: Out of memory trying to malloc space for >info< tag\n");
						return true;
					};
					tfile=fopen(tmp_ch->fname,"rb");
					if (!tfile) {
						fprintf(stderr,"ERROR: Could not open >info< tag file %s\n",tmp_ch->fname);
						PRINTERRCODE();
						return true;
					};
					if ( fread(cfg->info,st.st_size,1,tfile) != 1 ) {
						fprintf(stderr,"ERROR: Could not read >info< tag file %s\n",tmp_ch->fname);
						PRINTERRCODE();
						return true;
					};
					cfg->info[st.st_size]=0;
//					while ((cfg->info[strlen(cfg->info)]== 0x0A)||
//							(cfg->info[strlen(cfg->info)]== 0x0D)) 
//						cfg->info[strlen(cfg->info)]=0;
					fprintf(stdout,"Info length: %d\n",strlen(cfg->info));
					fprintf(stdout,"Info: %s\n",cfg->info);
					break;
				case CHUNK_WHAT:
					cfg->what_read=true;
					if (STAT(tmp_ch->fname,&st)) {
						fprintf(stderr,"ERROR: Could not stat file %s\n",tmp_ch->fname);
						return true;
					};
					if (st.st_size==0)
						fprintf(stderr,"Do you really intend not to set the >what< tag?\n");
					cfg->what = (char*)(malloc(st.st_size+1));
					fprintf(stdout,"what file: %s\t size: %d\n",tmp_ch->fname,st.st_size);
					memchr(cfg->what,0,st.st_size+1);
					if (!(cfg->what)) {
						fprintf(stderr,"ERROR: Out of memory trying to malloc space for >what< tag\n");
						return true;
					};
					tfile=fopen(tmp_ch->fname,"rb");
					if (!tfile) {
						fprintf(stderr,"ERROR: Could not open >what< tag file %s\n",tmp_ch->fname);
						PRINTERRCODE();
						return true;
					};
					if ( fread(cfg->what,st.st_size,1,tfile) != 1 ) {
						fprintf(stderr,"ERROR: Could not read >what< tag file %s\n",tmp_ch->fname);
						PRINTERRCODE();
						return true;
					};
					cfg->what[st.st_size]=0;
//					while ((cfg->what[strlen(cfg->what)]== 0x0A)||
//							(cfg->what[strlen(cfg->what)]== 0x0D)) 
//						cfg->what[strlen(cfg->what)]=0;
					fprintf(stdout,"What length: %d\n",strlen(cfg->what));
					fprintf(stdout,"What: %s\n",cfg->what);
					break;
				case CHUNK_RELEASE:
					cfg->rel_read=true;
					if (STAT(tmp_ch->fname,&st)) {
						fprintf(stderr,"ERROR: Could not stat file %s\n",tmp_ch->fname);
						return true;
					};
					if (st.st_size==0)
						fprintf(stderr,"Do you really intend not to set the release tag?\n");
					cfg->release = (char*)(malloc(st.st_size+1));
					fprintf(stdout,"Release file: %s\t size: %d\n",tmp_ch->fname,st.st_size);
					memchr(cfg->release,0,st.st_size+1);
					if (!(cfg->release)) {
						fprintf(stderr,"ERROR: Out of memory trying to malloc space for release tag\n");
						return true;
					};
					tfile=fopen(tmp_ch->fname,"rb");
					if (!tfile) {
						fprintf(stderr,"ERROR: Could not open release tag file %s\n",tmp_ch->fname);
						PRINTERRCODE();
						return true;
					};
					if ( fread(cfg->release,st.st_size,1,tfile) != 1 ) {
						fprintf(stderr,"ERROR: Could not read >release< tag file %s\n",tmp_ch->fname);
						PRINTERRCODE();
						return true;
					};
					cfg->release[st.st_size]=0;
//					while ((cfg->release[strlen(cfg->release)]== 0x0A)||
//							(cfg->release[strlen(cfg->release)]== 0x0D)) 
//						cfg->release[strlen(cfg->release)]=0;
					fprintf(stdout,"Release length: %d\n",strlen(cfg->release));
					fprintf(stdout,"Release: %s\n",cfg->release);
					break;
				case CHUNK_VERSION:	
					cfg->ver_read=true;
					if (STAT(tmp_ch->fname,&st)) {
						fprintf(stderr,"ERROR: Could not stat file %s\n",tmp_ch->fname);
						return true;
					};
					if (st.st_size==0)
						fprintf(stderr,"Do you really intend not to set the version tag?\n");
					cfg->version = (char*)(malloc(st.st_size+1));
					fprintf(stdout,"Version file: %s\t size: %d\n",tmp_ch->fname,st.st_size);
					memchr(cfg->version,0,st.st_size+1);
					if (!(cfg->version)) {
						fprintf(stderr,"ERROR: Out of memory trying to malloc space for version tag\n");
						return true;
					};
					tfile=fopen(tmp_ch->fname,"rb");
					if (!tfile) {
						fprintf(stderr,"ERROR: Could not open version tag file\n");
						PRINTERRCODE();
						return true;
					};
					if ( fread(cfg->version,st.st_size,1,tfile) != 1 ) {
						fprintf(stderr,"ERROR: Could not read >version< tag file %\n",tmp_ch->fname);
						PRINTERRCODE();
						return true;
					};
					cfg->version[st.st_size]=0;
//					while ((cfg->version[strlen(cfg->version)]== 0x0A)||
//							(cfg->version[strlen(cfg->version)]== 0x0D)) 
//						cfg->version[strlen(cfg->version)]=0;
					fprintf(stdout,"Version length: %d\n",strlen(cfg->version));
					fprintf(stdout,"Version: %s\n",cfg->version);
					break;
				case CHUNK_HWREV:	
					cfg->ver_read=true;
					if (STAT(tmp_ch->fname,&st)) {
						fprintf(stderr,"ERROR: Could not stat file %s\n",tmp_ch->fname);
						return true;
					};
					if (st.st_size==0)
						fprintf(stderr,"Do you really intend not to set the hardware revision tag?\n");
					cfg->hwrev = (char*)(malloc(st.st_size+1));
					fprintf(stdout,"Version file: %s\t size: %d\n",tmp_ch->fname,st.st_size);
					memchr(cfg->hwrev,0,st.st_size+1);
					if (!(cfg->hwrev)) {
						fprintf(stderr,"ERROR: Out of memory trying to malloc space for hardware revision tag\n");
						return true;
					};
					tfile=fopen(tmp_ch->fname,"rb");
					if (!tfile) {
						fprintf(stderr,"ERROR: Could not open hardware revision tag file\n");
						PRINTERRCODE();
						return true;
					};
					if ( fread(cfg->hwrev,st.st_size,1,tfile) != 1 ) {
						fprintf(stderr,"ERROR: Could not read >hwrev< tag file %\n",tmp_ch->fname);
						PRINTERRCODE();
						return true;
					};
					cfg->hwrev[st.st_size]=0;
//					while ((cfg->hwrev[strlen(cfg->hwrev)]== 0x0A)||
//							(cfg->hwrev[strlen(cfg->hwrev)]== 0x0D)) 
//						cfg->hwrev[strlen(cfg->hwrev)]=0;
					fprintf(stdout,"Hardware revision tag length: %d\n",strlen(cfg->hwrev));
					fprintf(stdout,"Hardware revision tag: %s\n",cfg->hwrev);
					break;
				default:	fprintf(stderr,"ERROR: This error should not occure under any (even faulty) circumstances\n");
							fprintf(stderr,"       In other words: The code is really fucked up if this error occures\n");
							return true;
			}; // switch(tcp_ch->typ)
			// leave cur_idx1 as it is:
			// that chunk was replaced by the yet unhandled chunk at cur_idx1+1
		} else { //if (chunktype neither INFO, WHAT, RELEASE nor VERSION)
			cchnk=cfg->chnk[cur_idx1];
			ctyp=cchnk->typ;
			if ((ctyp == CHUNK_PUMPHDA) 
			  || (ctyp == CHUNK_PUMPHDB)
			  || (ctyp == CHUNK_PUMPHDC)) {
				cfg->upgrade_size+=8;				// type and length only
				cchnk->fname=(char*)NULL;
			} else {
				if ((cchnk->typ==CHUNK_FLASHLOADER)
				  ||(cchnk->typ==CHUNK_FLASHKERNEL)
				  ||(cchnk->typ==CHUNK_FLASHRAMDISK)
				  ||(cchnk->typ==CHUNK_FLASHRANDOM)) {
					cchnk->length=4;	// flash commands need to write the address
				} else {
					cchnk->length=0;	// any others don't
				};
				if (STAT(cchnk->fname,&st)) {
					fprintf(stderr,"ERROR: Could not stat file %s\n",cchnk->fname);
					return true;
				};
				if ((st.st_mode & _S_IFREG) != _S_IFREG) {
					fprintf(stderr,"ERROR: %s is not a regular file\n",cchnk->fname);
					return true;
				};
				cchnk->length+=st.st_size;
				cfg->upgrade_size+=8+cchnk->length; // type, length and the data itself
			};
			fprintf(stdout,"Chunk %d:\n",cur_idx1+1);
			fprintf(stdout,"Chunk type:     0x%2.2X\n",cchnk->typ);
			fprintf(stdout,"Chunk filename: %s\n",cchnk->fname);
			fprintf(stdout,"Chunk address:  0x%8.8X\n",cchnk->addr);
			fprintf(stdout,"Chunk size: %d (including address if necessary)\n",cchnk->length);
			cur_idx1++;
		}; // end of:  if (chunktype either INFO, WHAT, RELEASE or VERSION)
		fprintf(stdout,"length of upgrade now: %d\n",cfg->upgrade_size);
	}; // end of:  while (cur_idx1 < cfg->nchunks)
	if (cfg->nchunks==0) {
		fprintf(stderr,"ERROR: No chunk that would be transfered to the empeg/RioCar was defined\n");
		return true;
	};
	if (!cfg->info)
		fprintf(stderr,"WARNING: No info text set\n");
	if (!cfg->what_read)
		fprintf(stderr,"WARNING: No \"what\" text set\n");
	if (!cfg->ver_read)
		fprintf(stderr,"WARNING: No version text set\n");
	if (!cfg->rel_read)
		fprintf(stderr,"WARNING: No release text set\n");
	if ((!cfg->info_read)&&(!cfg->what_read)&&(!cfg->rel_read)&&(!cfg->ver_read)) {
		fprintf(stderr,"WARNING: There is no informative text set at all, so I will set some defaults\n");
		cfg->info="Build by UpgradeBuilder (c)2001 Sven Mueller <sven@incase.de>\n";
		cfg->what="an upgrade file\n";
		cfg->release="\n";
		cfg->version="\n";
	};
	// we need to update the upgrade_size variable to include the header fields as well
	cfg->upgrade_size += 8+4;//sizeof(cfg->upgraderversion)	// CHUNK_UPGRADERVERSION
	cfg->upgrade_size += 8+strlen(cfg->info);
	cfg->upgrade_size += 8+strlen(cfg->what);
	cfg->upgrade_size += 8+strlen(cfg->version);
	cfg->upgrade_size += 8+strlen(cfg->release);
	cfg->upgrade_size += 8+strlen(cfg->hwrev);
	fprintf(stdout,"Total upgrade data size: %d bytes\n",cfg->upgrade_size);
	fprintf(stdout,"Final upgrade file will contain %d bytes\n",cfg->upgrade_size+8);
	if (!cfg->doit)
		fprintf(stdout,"INFO: not writing an upgrade file\n");
	// everything within limits.
	return false;
};

bool writeUL(FILE* f, const unsigned long ulong) {
	unsigned char fil[4];
	unsigned long ul;

	ul=ulong;
	fil[0]=(unsigned char)(ul & 0xFF);
	ul>>=8;
	fil[1]=(unsigned char)(ul & 0xFF);
	ul>>=8;
	fil[2]=(unsigned char)(ul & 0xFF);
	ul>>=8;
	fil[3]=(unsigned char)(ul & 0xFF);
	if (fwrite((void*) fil,4,1,f) != 1) {
		fprintf(stderr,"ERROR: Could not write ulong\n");
		return true;
	};
	return false;
};


bool write_file(const config *cfg) {
	FILE* f;
	int i;
	fprintf(stdout,"Writing upgrade file now\n");
	f=fopen(cfg->outname,"wb");
	if (!f) {
		fprintf(stderr,"ERROR: Could not open upgrade file >%s< for output\n",cfg->outname);
		PRINTERRCODE();
		return true;
	};
	if (writeUL(f,cfg->upgrade_size)
		|| writeUL(f,CHUNK_UPGRADERVERSION) 
		|| writeUL(f,sizeof(unsigned long))
		|| writeUL(f,cfg->upgraderversion)) {
		fprintf(stderr,"ERROR: Could not write to upgrade file\n");
		PRINTERRCODE();
		return true;
	};
	fprintf(stdout,"Wrote file size info and Upgrader version chunk.\n");
	if (writeUL(f,CHUNK_INFO)
		|| writeUL(f,strlen(cfg->info))) {
		fprintf(stderr,"ERROR: Could not write to upgrade file\n");
		PRINTERRCODE();
		return true;
	};
	if (fwrite(cfg->info,strlen(cfg->info),1,f)!=1) {
		fprintf(stderr,"ERROR: Could not write to upgrade file\n");
		PRINTERRCODE();
		return true;
	};
	fprintf(stdout,"Wrote info chunk.\n");
	if (writeUL(f,CHUNK_WHAT)
		|| writeUL(f,strlen(cfg->what))) {
		fprintf(stderr,"ERROR: Could not write to upgrade file\n");
		PRINTERRCODE();
		return true;
	};
	if (fwrite(cfg->what,strlen(cfg->what),1,f)!=1) {
		fprintf(stderr,"ERROR: Could not write to upgrade file\n");
		PRINTERRCODE();
		return true;
	};
	fprintf(stdout,"Wrote \"what\" chunk.\n");
	if (writeUL(f,CHUNK_RELEASE)
		|| writeUL(f,strlen(cfg->release))) {
		fprintf(stderr,"ERROR: Could not write to upgrade file\n");
		PRINTERRCODE();
		return true;
	};
	if (fwrite(cfg->release,strlen(cfg->release),1,f)!=1) {
		fprintf(stderr,"ERROR: Could not write to upgrade file\n");
		PRINTERRCODE();
		return true;
	};
	fprintf(stdout,"Wrote release chunk.\n");
	if (writeUL(f,CHUNK_VERSION)
		|| writeUL(f,strlen(cfg->version))) {
		fprintf(stderr,"ERROR: Could not write to upgrade file\n");
		PRINTERRCODE();
		return true;
	};
	if (fwrite(cfg->version,strlen(cfg->version),1,f)!=1) {
		fprintf(stderr,"ERROR: Could not write to upgrade file\n");
		PRINTERRCODE();
		return true;
	};
	fprintf(stdout,"Wrote version chunk.\n");
	if (writeUL(f,CHUNK_HWREV)
		|| writeUL(f,strlen(cfg->hwrev))) {
		fprintf(stderr,"ERROR: Could not write to upgrade file\n");
		PRINTERRCODE();
		return true;
	};
	if (fwrite(cfg->hwrev,strlen(cfg->hwrev),1,f)!=1) {
		fprintf(stderr,"ERROR: Could not write to upgrade file\n");
		PRINTERRCODE();
		return true;
	};
	fprintf(stdout,"Wrote hwrev chunk.\n");
	fprintf(stdout,"Now commencing with content chunks.\n");
	for (i=0; i<cfg->nchunks; i++) {
		if (writeUL(f,cfg->chnk[i]->typ)
		  || writeUL(f,cfg->chnk[i]->length)) {
			fprintf(stderr,"ERROR: Could not write to upgrade file\n");
			PRINTERRCODE();
			return true;
		};
		switch (cfg->chnk[i]->typ) {
			case CHUNK_FLASHLOADER:
			case CHUNK_FLASHKERNEL:
			case CHUNK_FLASHRAMDISK:
			case CHUNK_FLASHRANDOM:
				fprintf(stdout,"Writing a flash chunk\n");
				if (writeUL(f,cfg->chnk[i]->addr)) {
					fprintf(stderr,"ERROR: Could not write to upgrade file\n");
					PRINTERRCODE();
					return true;
				};
				if (appendfile(f,cfg->chnk[i]->fname)) {
					fprintf(stderr,"ERROR: Could not append chunk file %s to upgrade file\n",cfg->chnk[i]->fname);
					PRINTERRCODE();
					return true;
				};
				break;
			case CHUNK_PUMPHDA:
			case CHUNK_PUMPHDB:
			case CHUNK_PUMPHDC:
				fprintf(stdout,"Writing a partitioning chunk\n");
				// already done, this only contains the chunk header with no data
				break;
			case CHUNK_PUMPHDA1:
			case CHUNK_PUMPHDA2:
			case CHUNK_PUMPHDA3:
			case CHUNK_PUMPHDA4:
			case CHUNK_PUMPHDA5:
			case CHUNK_PUMPHDA6:
			case CHUNK_PUMPHDA7:
			case CHUNK_PUMPHDA8:
			case CHUNK_PUMPHDB1:
			case CHUNK_PUMPHDB2:
			case CHUNK_PUMPHDB3:
			case CHUNK_PUMPHDB4:
			case CHUNK_PUMPHDB5:
			case CHUNK_PUMPHDB6:
			case CHUNK_PUMPHDB7:
			case CHUNK_PUMPHDB8:
			case CHUNK_PUMPHDC1:
			case CHUNK_PUMPHDC2:
			case CHUNK_PUMPHDC3:
			case CHUNK_PUMPHDC4:
			case CHUNK_PUMPHDC5:
			case CHUNK_PUMPHDC6:
			case CHUNK_PUMPHDC7:
			case CHUNK_PUMPHDC8:
				fprintf(stdout,"Writing a partition image chunk\n");
				if (appendfile(f,cfg->chnk[i]->fname)) {
					fprintf(stderr,"ERROR: Could not append chunk file %s to upgrade file\n",cfg->chnk[i]->fname);
					PRINTERRCODE();
					return true;
				};
				break;
			default:
				fprintf(stderr,"Unknown chunk type: 0x%2.2X\n",cfg->chnk[i]->typ);
				return true;
		};
	};	
	fclose(f);
	return false;
};

bool appendfile(FILE* target, char* sourcename) {
	char* buf;
	FILE* sourcef;
	struct STAT st;
	OFF_T ctpos;
	OFF_T readsize;

	STAT(sourcename,&st);
	ctpos=0;
	buf=(char*)malloc(4096);
	if (!buf) {
		fprintf(stderr,"ERROR: Out of memory trying to append chunk file %s\n",sourcename);
		PRINTERRCODE();
		return true;
	};
	sourcef=fopen(sourcename,"rb");
	if (!sourcef) {
		fprintf(stderr,"ERROR: Could not open chunk file %s\n",sourcename);
		PRINTERRCODE();
		return true;
	};
	while (ctpos < st.st_size) {
		readsize=((ctpos+4096)<=st.st_size)?4096:(st.st_size-ctpos);
		if (fread(buf,readsize,1,sourcef)!=1) {
			fprintf(stderr,"ERROR: Could not read all data from chunk file %s\n",sourcename);
			PRINTERRCODE();
			return true;
		};
		if (fwrite(buf,readsize,1,target)!=1) {
			fprintf(stderr,"ERROR: Could not append to upgrade file, reading from chunk file %s\n",sourcename);
			PRINTERRCODE();
			return true;
		};
		ctpos += readsize;
	};
	fclose(sourcef);
	return false;
};

bool do_crc32(const config *cfg) {
	char*			buf;
	unsigned long	mycrc;
	OFF_T			curpos;
	OFF_T			readsize;
	FILE*			f;

	f=fopen(cfg->outname,"rb");
	if (!f) {
		fprintf(stderr,"ERROR: Could not open upgrade file to calculate CRC32 of it\n");
		PRINTERRCODE();
		return true;
	};
	fseek(f,4,SEEK_SET);	// skip over size info, we already now that
	curpos=ftell(f);
	if (curpos != 4) {
		fprintf(stderr,"ERROR: Either fseek or ftell went wrong while skipping over size info\n");
		PRINTERRCODE();
		return true;
	};
	buf=(char*)malloc(4096);
	if (!buf) {
		fprintf(stderr,"ERROR: Out of memory allocating buffer memory or CRC calculation\n");
		return true;
	};
	mycrc=0;
	while (curpos < (cfg->upgrade_size+4)) {
		readsize=((curpos+4096) < (cfg->upgrade_size+4))?4096:((cfg->upgrade_size+4)-curpos);
		if (fread(buf,readsize,1,f)!=1) {
			fprintf(stderr,"ERROR: could not read data from upgrade file for CRC calculation\n");
			PRINTERRCODE();
			return true;
		};
		mycrc=CRC::CRC32(mycrc,(unsigned char*)buf,readsize);
		curpos+=readsize;
	};
	fclose(f);
	f=fopen(cfg->outname,"ab");
	if (!f) {
		fprintf(stderr,"ERROR: Could not open upgrade file to append CRC to it\n");
		PRINTERRCODE();
		return true;
	};
	if (writeUL(f,mycrc)) {
		fprintf(stderr,"ERROR: Could not write CRC to upgrade file\n");
		PRINTERRCODE();
		return true;
	};
	fclose(f);
	return false;
};

