// UpgradeSplitter.cpp
// Split up empeg upgrade files into its chunks

#ifndef WIN32
#define _GNU_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#ifdef WIN32
#include <io.h>
#endif
#include "crc.h"
#include "defines.h"

#define BUFSIZE		4096

#ifdef WIN32
#define STAT _stat
#define OFF_T _off_t
#else
#define _S_IFREG S_IFREG
#define OFF_T __off_t
#define STAT stat
#endif

typedef struct configtag {
	bool	write_files;
	bool	show_help;
	char	inputfilename[256];
	char	outputfile[256];
	FILE	*infile;
	FILE	*ctlfile;
	long	curpos;
	long	inlength;
} config;

// turn an error number into its appropriate string
const char* errstr(int errnumber){
	static char	buf[256];

	switch (errnumber) {
		case	ENOENT:		return("No such file or directory"); 
							break;
		case	E2BIG:		return("Argument list too long");
							break;
		case	ENOEXEC:	return("Exec format error");
							break;
		case	EBADF:		return("Bad file number");
							break;
		case	ECHILD:		return("No spawned processes");
							break;
		case	EAGAIN:		return("No more processes or not enough memory or maximum nesting level reached");
							break;
		case	ENOMEM:		return("Not enough memory");
							break;
		case	EACCES:		return("Permission denied");  
							break;
		case	EEXIST:		return("File exists");
							break;
		case	EXDEV:		return("Cross-device link");
							break;
		case	EINVAL:		return("Invalid argument");
							break;
		case	EMFILE:		return("Too many open files");
							break;
		case	ENOSPC:		return("No space left on device");
							break;
		case	EDOM:		return("Math argument");
							break;
		case	ERANGE:		return("Result too large");
							break;
		case	EDEADLOCK:	return("Resource deadlock would occur");
							break;
		default:			sprintf(buf,"unknown error code: %d",errnumber);
							return(buf);
	}
};

void usage(FILE* f,char* exename){
	fprintf(f,"Usage:\n");
	fprintf(f,"%s [-h] [-f <outputfile>] [-w] <upgradefile>\n",exename);
	fprintf(f,"    -h              prints this help\n");
	fprintf(f,"    -f <outputfile> sets the name of the primary output file\n");
	fprintf(f,"                    defaults to \"upgrade.ctl\"\n");
	fprintf(f,"    -w              write the split up files\n");
	fprintf(f,"                    default: only print out structure of input file\n");
	fprintf(f,"    <upgradefile>   the name of the upgrade file to split/analyze\n");
	fprintf(f,"The upgrade file is split into various files, with a control file\n");
	fprintf(f,"(named upgrade.ctl) that lists the chunks in their original sequence\n");
	fprintf(f,"giving their filenames and other details\n");
};

bool checkargs(int argc, char* argv[], config* cfg) {
	int curarg=1;
	
	cfg->curpos=0;
	cfg->inputfilename[0]=0;
	strcpy(cfg->outputfile,"upgrade.ctl");
	cfg->show_help=false;
	cfg->write_files=false;

	while (curarg < argc) {
		if (argv[curarg][0]=='-') {
			switch(argv[curarg][1]) {
				case 'h':
				case 'H':	cfg->show_help=true;
							break;
				case 'w':
				case 'W':	cfg->write_files=true;
							break;
				case 'f':
				case 'F':	if (++curarg < argc) {
								strcpy(cfg->outputfile,argv[curarg]);
							} else {
								fprintf(stderr,"ERROR: Argument missing for -f option\n");
								return true;
							};
							break;
				default:	fprintf(stderr,"ERROR: Unknown command line option %s\n",argv[curarg]);
							usage(stderr,argv[0]);
							return true;
			};
		} else {
			strcpy(cfg->inputfilename,argv[curarg]);
		};
		curarg++;
	};
	return false;
};

unsigned long readUL (FILE *f) {
	unsigned char fil[4];
	unsigned long ul;

	fread((void*) fil,4,1,f);
	ul=(fil[0] | fil[1]<<8 | fil[2]<<16 | fil[3]<<24);
	return ul;
};

bool checkupfile(config* cfg) {
	unsigned long	cpos;	// current position in cfg.infile
	unsigned long	ccrc;	// current CRC32
	unsigned long	tlen;	// length of input file's data
	unsigned long	clen;	// length of current chunk
	unsigned char	*buf;	// input buffer
	unsigned long	fcrc;	// the claimed CRC32 of the input file
	unsigned long	blocktyp; // type of the current block

	buf=(unsigned char*) malloc(BUFSIZE);
	if (!buf) {
		fprintf(stderr,"ERROR: Out of memory while checking CRC\n");
		return true;
	};
	cpos=0;
	if (fseek(cfg->infile,cpos,SEEK_SET)) {
		fprintf(stderr,"ERROR: could not seek to beginning of input file\n");
		fprintf(stderr,"error code: %s\n",errstr(errno));
		return true;
	};
	
	tlen=readUL(cfg->infile);
	cfg->inlength=tlen;
	cpos += 4;

	// start CRC32 calculation with 0 (others might use 0xFFFFFFFF)
	ccrc=0;

	while ((cpos >= 0) && (cpos < (tlen+4))) {
		clen=((((tlen+4) - cpos) > BUFSIZE)?(BUFSIZE):((tlen+4)-cpos));
//		fprintf(stdout,"\rtlen: %8.8d cpos: %8.8d clen: %8.8d",tlen,cpos,clen);
		if (fread((void *) buf,1,clen,cfg->infile) != clen) {
			fprintf(stderr,"ERROR: Unexpected read error checking input file CRC\n");
			fprintf(stderr,"Error code: %s\n",errstr(errno));
			return true;
		};
		ccrc=CRC::CRC32(ccrc,buf,clen);
		cpos+=clen;
//		fprintf(stdout,"\r%8.8d",cpos);
	}; // while ((cpos >= 0) && (cpos < tlen))
	fcrc=readUL(cfg->infile);

	if (fcrc != ccrc) {
		fprintf(stderr,"ERROR: CRC32 does not match input file contents\n");
		return true;
	};

	// if the CRC is ok, walk through the file structure to check wether it is ok

	// reset position to start of first chunk
	cpos=4;

	while (cpos < (tlen+4)) {
		if (fseek(cfg->infile,cpos,SEEK_SET)) {
			fprintf(stderr,"ERROR: could not seek to start of next chunk while checking input file structure\n");
			fprintf(stderr,"Error code: %s\n",errstr(errno));
			return true;
		};
		blocktyp=readUL(cfg->infile);
		cpos+=4;
		clen=readUL(cfg->infile);
		cpos+=4;
		cpos+=clen;
	};
	// if everything went well, we should now have a cpos of 4+tlen
	if (cpos != (tlen+4)) {
		fprintf(stderr,"ERROR: Last chunk not within bounds\n");
		return true;
	};
	// all checks done
	return false;
};

char* getfreename(const char* basname){
	char*	buf;
	char*	rbuf;
	struct STAT st;
	size_t	curlen;
	bool	exists;
	static char	*posext="0123456789abcdefghijklmnopqrstuvwxyz";
	int		numexts;
	int		curext=0;

	buf=(char*)malloc(256);
	if (!buf) {
		fprintf(stderr,"ERROR: out of memory trying to find a new extenson\n");
	};
	strncpy(buf,basname,255);
	buf[255]=0;
	curlen=strlen(buf);
	buf[curlen]=0;
	buf[curlen+1]=0;
	numexts=strlen(posext);
	if (STAT(buf,&st)!=0) {
		exists=(errno=ENOENT)?false:true;
	};

	
	// if file "basname" does not exist: return it as a free name
	if (!exists) return buf; 
	
	// first check wether a single character extension is enough
	while (exists && (curext<numexts)) {
		buf[curlen]=posext[curext];
		if (STAT(buf,&st)!=0) {
			exists=(errno=ENOENT)?false:true;
		};
		if (!exists) return buf;
		curext++;
	};

	curext=0;
	while (exists && (curext<numexts)) {
		buf[curlen]=posext[curext];
		rbuf=getfreename(buf);
		if (rbuf) {
			free(buf);
			return rbuf;
		};
		curext++;
	};
	free(buf);
	return((char*)NULL);
};

// write <size> bytes from buf* to filename fnam
bool writechunk(const unsigned long size, const char* buf, const char* fnam, config *cfg) {
	FILE *f;
	f=fopen(fnam,"wb");
	if (!f) {
		fprintf(stderr,"ERROR: Could not open file %s\n",fnam);
		fprintf(stderr,"Error code: %s\n",errstr(errno));
		return true;
	};
	if (fwrite(buf,size,1,f)!=1) {
		fprintf(stderr,"ERROR: Could not write to file %s\n",fnam);
		fprintf(stderr,"Error code: %s\n",errstr(errno));
		return true;
	};
	fclose(f);
	return false;
};

// copy <size> bytes from cfg->infile to filename fnam
bool copychunk(const unsigned long size, const char *fnam, config *cfg) {
	unsigned long count=0;
	unsigned long s=0;
	char* buf;
	FILE* f;

	f=fopen(fnam,"wb");
	if (!f) {
		fprintf(stderr,"ERROR: Could not open file %s\n",fnam);
		fprintf(stderr,"Error code: %s\n",errstr(errno));
		return true;
	};
	buf=(char*)malloc(BUFSIZE);
	while (count < size) {
		s=(size-count < BUFSIZE)?(size-count):BUFSIZE;
		if (fread(buf,s,1,cfg->infile)!=1) {
			fprintf(stderr,"ERROR: Could not read from input file %s\n",cfg->inputfilename);
			fprintf(stderr,"Error code: %s\n",errstr(errno));
			return true;
		};
		if (fwrite(buf,s,1,f)!=1) {
			fprintf(stderr,"ERROR: Could not write to file %s\n",fnam);
			fprintf(stderr,"Error code: %s\n",errstr(errno));
			return true;
		};
		count+=s;
	};
	fclose(f);
	return false;
};

int main(int argc, char* argv[])
{
	config			cfg;
	unsigned long	chunklen;
	unsigned long	chunktyp;
	unsigned long	upgver		=2;
	char			*info		=NULL;
	char			*what		=NULL;
	char			*release	=NULL;
	char			*version	=NULL;
	char			*hwrev		=NULL;
	char			*fnam		=NULL;
	char			*tnam		=NULL;
	unsigned long	flashaddr	=0;
	
	if (argc<2) {
		fprintf(stderr,"ERROR: To few arguments\n");
		usage(stderr,argv[0]);
		return 1;
	};

	// check commandline arguments and set the
	if (checkargs(argc,argv,&cfg)) return 1;
	
	// did the user ask for usage help?
	if (cfg.show_help) {
		// yes: print usage info and exit
		usage(stdout,argv[0]);
		return 0;
	};

	// open the input file
	cfg.infile=fopen(cfg.inputfilename,"rb");
	if (cfg.infile == NULL) {
		fprintf(stderr,"Can't open inputfile <%s>\n",cfg.inputfilename);
		fprintf(stderr,"Error code: %s\n",errstr(errno));
		return 1;
	};

	// open the output control file if necessary
	if (cfg.write_files) {
		cfg.ctlfile=fopen(cfg.outputfile,"wt");
		if (cfg.ctlfile == NULL) {
			fprintf(stderr,"Can't open outputfile <%s>\n",cfg.outputfile);
			fprintf(stderr,"Error code: %s\n",errstr(errno));
			return 1;
		};
	};

	// check the input file for consistency (CRC32 check)
	// exit if it fails the test
	if (checkupfile(&cfg)) return 1;

	// go to the first chunk
	if (fseek(cfg.infile,4,SEEK_SET)) {
		fprintf(stderr,"Error seeking to first chunk\n");
		fprintf(stderr,"Error code: %s\n",errstr(errno));
		fcloseall();
		return 1;
	};
	cfg.curpos=4;

	// now read the chunks, one by one and handle them appropriately
	while ((cfg.curpos < cfg.inlength) && (cfg.curpos != 0)) {
		fseek(cfg.infile,cfg.curpos,SEEK_SET);
		chunktyp = readUL(cfg.infile);
		chunklen = readUL(cfg.infile);
		cfg.curpos+=8;
		switch(chunktyp) {
			case CHUNK_UPGRADERVERSION:		
				upgver=readUL(cfg.infile);
				if (upgver>UPGRADERVERSION) {
					fprintf(stderr,"ERROR: UpgradeSplitter has an older version (%d) than upgrade file (%d)\n",UPGRADERVERSION,upgver);
					return 1;
				};
				if (cfg.write_files) fprintf(cfg.ctlfile,"upgraderversion=%d\n",upgver);
				fprintf(stdout,"upgraderversion=%#2.2x\n",upgver);
				break;
			case CHUNK_INFO:
				info=(char*)malloc(chunklen+1);
				if (!info) {
					fprintf(stderr,"ERROR: out of memory reading \"info\" chunk\n");
					return 1;
				};
				if (fread((void*)info,chunklen,1,cfg.infile)!=1) {
					fprintf(stderr,"ERROR: Could not read \"info\" chunk\n");
					fprintf(stderr,"Error code: %s\n",errstr(errno));
					return 1;
				};
				info[chunklen]=0;
				fprintf(stdout,"info=%s",info);
				if (cfg.write_files) {
					if (fnam) free(fnam);
					fnam=getfreename("upg_infofile");
					if (!fnam) {
						fprintf(stderr,"ERROR while finding free filename for \"info\" chunk\n");
						return 1;
					};
					fprintf(cfg.ctlfile,"infofile=%s\n",fnam);
					if (writechunk(chunklen,info,fnam,&cfg)) {
						fprintf(stderr,"ERROR writing \"info\" chunk\n");
						return 1;
					};
				};
				break;
			case CHUNK_WHAT:
				what=(char*)malloc(chunklen+1);
				if (!what) {
					fprintf(stderr,"ERROR: out of memory reading \"what\" chunk\n");
					return 1;
				};
				if (fread((void*)what,chunklen,1,cfg.infile)!=1) {
					fprintf(stderr,"ERROR: Could not read \"what\" chunk\n");
					fprintf(stderr,"Error code: %s\n",errstr(errno));
					return 1;
				};
				what[chunklen]=0;
				fprintf(stdout,"what=%s",what);
				if (cfg.write_files) {
					if (fnam) free(fnam);
					fnam=getfreename("upg_whatfile");
					if (!fnam) {
						fprintf(stderr,"ERROR while finding free filename for \"what\" chunk\n");
						return 1;
					};
					fprintf(cfg.ctlfile,"whatfile=%s\n",fnam);
					if (writechunk(chunklen,what,fnam,&cfg)) {
						fprintf(stderr,"ERROR writing \"what\" chunk\n");
						return 1;
					};
				};
				break;
			case CHUNK_RELEASE:
				release=(char*)malloc(chunklen+1);
				if (!release) {
					fprintf(stderr,"ERROR: out of memory reading \"release\" chunk\n");
					return 1;
				};
				if (fread((void*)release,chunklen,1,cfg.infile)!=1) {
					fprintf(stderr,"ERROR: Could not read \"release\" chunk\n");
					fprintf(stderr,"Error code: %s\n",errstr(errno));
					return 1;
				};
				release[chunklen]=0;
				fprintf(stdout,"release=%s",release);
				if (cfg.write_files) {
					if (fnam) free(fnam);
					fnam=getfreename("upg_releasefile");
					if (!fnam) {
						fprintf(stderr,"ERROR while finding free filename for \"release\" chunk\n");
						return 1;
					};
					fprintf(cfg.ctlfile,"releasefile=%s\n",fnam);
					if (writechunk(chunklen,release,fnam,&cfg)) {
						fprintf(stderr,"ERROR writing \"release\" chunk\n");
						return 1;
					};
				};
				break;
			case CHUNK_VERSION:
				version=(char*)malloc(chunklen+1);
				if (!version) {
					fprintf(stderr,"ERROR: out of memory reading \"version\" chunk\n");
					return 1;
				};
				if (fread((void*)version,chunklen,1,cfg.infile)!=1) {
					fprintf(stderr,"ERROR: Could not read \"version\" chunk\n");
					fprintf(stderr,"Error code: %s\n",errstr(errno));
					return 1;
				};
				version[chunklen]=0;
				fprintf(stdout,"version=%s",version);
				if (cfg.write_files) {
					if (fnam) free(fnam);
					fnam=getfreename("upg_versionfile");
					if (!fnam) {
						fprintf(stderr,"ERROR while finding free filename for \"version\" chunk\n");
						return 1;
					};
					fprintf(cfg.ctlfile,"versionfile=%s\n",fnam);
					if (writechunk(chunklen,version,fnam,&cfg)) {
						fprintf(stderr,"ERROR writing \"version\" chunk\n");
						return 1;
					};
				};
				break;
			case CHUNK_FLASHLOADER:
			case CHUNK_FLASHKERNEL:
			case CHUNK_FLASHRAMDISK:
			case CHUNK_FLASHRANDOM:
				flashaddr=readUL(cfg.infile);
				if (fnam) free(fnam);
				switch (chunktyp) {
					case CHUNK_FLASHLOADER:
						fprintf(stdout,"flashloader address %#8.8x\n",flashaddr);
						if (cfg.write_files) {
							fnam=getfreename("upg_flashloader");
							if (!fnam) {
								fprintf(stderr,"ERROR while finding free filename for \"flashloader\" chunk\n");
								return 1;
							};
							fprintf(cfg.ctlfile,"flashloader=");
						};
						break;
					case CHUNK_FLASHKERNEL:
						fprintf(stdout,"flashkernel address %#8.8x\n",flashaddr);
						if (cfg.write_files) {
							fnam=getfreename("upg_flashkernel");
							if (!fnam) {
								fprintf(stderr,"ERROR while finding free filename for \"flashkernel\" chunk\n");
								return 1;
							};
							fprintf(cfg.ctlfile,"flashkernel=");
						};
						break;
					case CHUNK_FLASHRAMDISK:
						fprintf(stdout,"flashramdisk address %#8.8x\n",flashaddr);
						if (cfg.write_files) {
							fnam=getfreename("upg_flashramdisk");
							if (!fnam) {
								fprintf(stderr,"ERROR while finding free filename for \"flashramdisk\" chunk\n");
								return 1;
							};
							fprintf(cfg.ctlfile,"flashramdisk=");
						};
						break;
					case CHUNK_FLASHRANDOM:
						fprintf(stdout,"flashrandom address %#8.8x\n",flashaddr);
						if (cfg.write_files) {
							fnam=getfreename("upg_flashrandom");
							if (!fnam) {
								fprintf(stderr,"ERROR while finding free filename for \"flashrandom\" chunk\n");
								return 1;
							};
							fprintf(cfg.ctlfile,"flashkernel=");
						};
						break;
				};
				if(cfg.write_files) {
					if (copychunk(chunklen-4,fnam,&cfg)) {
						fprintf(stderr,"ERROR while copying flash chunk %s\n",fnam);
						return 1;
					};
					fprintf(cfg.ctlfile,"%#8.8x,%s\n",flashaddr,fnam);
				};
				break;
			case CHUNK_PUMPHDA:
			case CHUNK_PUMPHDB:
			case CHUNK_PUMPHDC:
				if (chunklen!=0) {
						fprintf(stderr,"ERROR: unexpected non-zero chunk size for partition table initialization\n");
						return 1;
				};
				// the "=1" following the pumphd[abc] string is a dummy argument that makes
				// parsing the control file much easier
				fprintf(stdout,"pumphd%c=1\n",(chunktyp-CHUNK_PUMPHDA+'a'));
				if (cfg.write_files) 
					fprintf(cfg.ctlfile,"pumphd%c\n",(chunktyp-CHUNK_PUMPHDA+'a'));
				break;
			case CHUNK_PUMPHDA1:
			case CHUNK_PUMPHDA2:
			case CHUNK_PUMPHDA3:
			case CHUNK_PUMPHDA4:
			case CHUNK_PUMPHDA5:
			case CHUNK_PUMPHDA6:
			case CHUNK_PUMPHDA7:
			case CHUNK_PUMPHDA8:
			case CHUNK_PUMPHDB1:
			case CHUNK_PUMPHDB2:
			case CHUNK_PUMPHDB3:
			case CHUNK_PUMPHDB4:
			case CHUNK_PUMPHDB5:
			case CHUNK_PUMPHDB6:
			case CHUNK_PUMPHDB7:
			case CHUNK_PUMPHDB8:
			case CHUNK_PUMPHDC1:
			case CHUNK_PUMPHDC2:
			case CHUNK_PUMPHDC3:
			case CHUNK_PUMPHDC4:
			case CHUNK_PUMPHDC5:
			case CHUNK_PUMPHDC6:
			case CHUNK_PUMPHDC7:
			case CHUNK_PUMPHDC8:
				fprintf(stdout,"pumphd%c%c with a %#8.8x sized compressed image\n",
						(((chunktyp & 0x70)>>4)-(CHUNK_PUMPHDA1>>4)+'a'),
						((chunktyp & 0x0f)-(CHUNK_PUMPHDA1 & 0x0f)+'1'),
						chunklen);
				if (cfg.write_files) {
					if (tnam) free(tnam);
					tnam=(char*)malloc(256);
					if (!tnam) {
						fprintf(stderr,"ERROR: out of memory while creating filename for partion image chunk\n");
						return 1;
					};
					sprintf(tnam,"upg_pumphd%c%c",
						(((chunktyp & 0x70)>>4)-(CHUNK_PUMPHDA1>>4)+'a'),
						((chunktyp & 0x0f)-(CHUNK_PUMPHDA1 & 0x0f)+'1'));
					fnam=getfreename(tnam);
					free(tnam);
					if (copychunk(chunklen,fnam,&cfg)) {
						fprintf(stderr,"ERROR while copying partition image chunk\n");
						return 1;
					};
					fprintf(cfg.ctlfile,"pumphd%c%c=%s\n",
						(((chunktyp & 0x70)>>4)-(CHUNK_PUMPHDA1>>4)+'a'),
						((chunktyp & 0x0f)-(CHUNK_PUMPHDA1 & 0x0f)+'1'),
						fnam);
				};
				break;
			case CHUNK_HWREV:	
				hwrev=(char*)malloc(chunklen+1);
				if (!hwrev) {
					fprintf(stderr,"ERROR: out of memory reading \"hwrev\" chunk\n");
					return 1;
				};
				if (fread((void*)hwrev,chunklen,1,cfg.infile)!=1) {
					fprintf(stderr,"ERROR: Could not read \"hwrev\" chunk\n");
					fprintf(stderr,"Error code: %s\n",errstr(errno));
					return 1;
				};
				hwrev[chunklen]=0;
				fprintf(stdout,"hwrev=%s",hwrev);
				if (cfg.write_files) {
					if (fnam) free(fnam);
					fnam=getfreename("upg_hwrevfile");
					if (!fnam) {
						fprintf(stderr,"ERROR while finding free filename for \"hwrev\" chunk\n");
						return 1;
					};
					fprintf(cfg.ctlfile,"hwrevfile=%s\n",fnam);
					if (writechunk(chunklen,hwrev,fnam,&cfg)) {
						fprintf(stderr,"ERROR writing \"hwrev\" chunk\n");
						return 1;
					};
				};
				break;
			default:	
				fprintf(stderr,"ERROR: Unknown chunk type %#2.2x found\n",chunktyp);
				return 1;
				break;
		}; // switch(chunktyp)
		cfg.curpos+=chunklen;
	}; // while ((cfg.curpos < cfg.inlength) && (cfg.curpos != 0))

	return 0;
}
